"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const glob_1 = __importDefault(require("glob"));
const os_1 = __importDefault(require("os"));
const tree_sitter_1 = __importDefault(require("tree-sitter"));
const tree_sitter_elm_1 = __importDefault(require("tree-sitter-elm"));
const vscode_uri_1 = require("vscode-uri");
const capabilityCalculator_1 = require("./capabilityCalculator");
const forest_1 = require("./forest");
const imports_1 = require("./imports");
const astProvider_1 = require("./providers/astProvider");
const codeActionProvider_1 = require("./providers/codeActionProvider");
const codeLensProvider_1 = require("./providers/codeLensProvider");
const completionProvider_1 = require("./providers/completionProvider");
const definitionProvider_1 = require("./providers/definitionProvider");
const diagnosticsProvider_1 = require("./providers/diagnostics/diagnosticsProvider");
const elmAnalyseDiagnostics_1 = require("./providers/diagnostics/elmAnalyseDiagnostics");
const documentFormatingProvider_1 = require("./providers/documentFormatingProvider");
const documentSymbolProvider_1 = require("./providers/documentSymbolProvider");
const foldingProvider_1 = require("./providers/foldingProvider");
const hoverProvider_1 = require("./providers/hoverProvider");
const referencesProvider_1 = require("./providers/referencesProvider");
const renameProvider_1 = require("./providers/renameProvider");
const workspaceSymbolProvider_1 = require("./providers/workspaceSymbolProvider");
const documentEvents_1 = require("./util/documentEvents");
const utils = __importStar(require("./util/elmUtils"));
const settings_1 = require("./util/settings");
const textDocumentEvents_1 = require("./util/textDocumentEvents");
class Server {
    constructor(connection, params) {
        this.calculator = new capabilityCalculator_1.CapabilityCalculator(params.capabilities);
        const forest = new forest_1.Forest();
        const imports = new imports_1.Imports();
        const parser = new tree_sitter_1.default();
        try {
            parser.setLanguage(tree_sitter_elm_1.default);
        }
        catch (error) {
            connection.console.info(error.toString());
        }
        const elmWorkspaceFallback = 
        // Add a trailing slash if not present
        params.rootUri && params.rootUri.replace(/\/?$/, "/");
        const elmWorkspace = vscode_uri_1.URI.parse(params.initializationOptions.elmWorkspace || elmWorkspaceFallback);
        const settings = new settings_1.Settings(params.capabilities, params.initializationOptions);
        if (elmWorkspace) {
            connection.console.info(`initializing - folder: "${elmWorkspace}"`);
            this.registerProviders(connection, forest, elmWorkspace, imports, settings, parser);
        }
        else {
            connection.console.info(`No workspace.`);
        }
    }
    get capabilities() {
        return {
            capabilities: this.calculator.capabilities,
        };
    }
    initialize(connection, forest, elmWorkspace, imports, parser) {
        try {
            const path = `${elmWorkspace.fsPath}elm.json`;
            connection.console.info(`Reading elm.json from ${path}`);
            // Find elm files and feed them to tree sitter
            const elmJson = require(path);
            const type = elmJson.type;
            const elmFolders = [];
            let elmVersion = "";
            if (type === "application") {
                elmVersion = elmJson["elm-version"];
                const sourceDirs = elmJson["source-directories"];
                sourceDirs.forEach((folder) => __awaiter(this, void 0, void 0, function* () {
                    elmFolders.push({
                        path: elmWorkspace.fsPath + folder,
                        writable: true,
                    });
                }));
            }
            else {
                // Todo find a better way to do this
                elmVersion = elmJson["elm-version"];
                if (elmVersion.indexOf(" ") !== -1) {
                    elmVersion = elmVersion.substring(0, elmVersion.indexOf(" "));
                }
                elmFolders.push({
                    path: `${elmWorkspace.fsPath}src`,
                    writable: true,
                });
            }
            elmFolders.push({
                path: `${elmWorkspace.fsPath}tests`,
                writable: true,
            });
            connection.console.info(`${elmFolders.length} source-dirs found`);
            const elmHome = this.findElmHome();
            // TODO find a way to detect this
            const packagesRoot = `${elmHome}/${elmVersion}/package/`;
            const dependencies = type === "application"
                ? Object.assign({}, elmJson.dependencies.direct, elmJson["test-dependencies"].direct) : Object.assign({}, elmJson.dependencies, elmJson["test-dependencies"]);
            for (const key in dependencies) {
                if (dependencies.hasOwnProperty(key)) {
                    const maintainer = key.substring(0, key.indexOf("/"));
                    const packageName = key.substring(key.indexOf("/") + 1, key.length);
                    // We should probably parse the elm json of a dependency, at some point down the line
                    const pathToPackage = type === "application"
                        ? `${packagesRoot}${maintainer}/${packageName}/${dependencies[key]}/src`
                        : `${packagesRoot}${maintainer}/${packageName}/${dependencies[key].substring(0, elmVersion.indexOf(" "))}`;
                    elmFolders.push({ path: pathToPackage, writable: false });
                }
            }
            const elmFilePaths = this.findElmFilesInFolders(elmFolders);
            connection.console.info(`Found ${elmFilePaths.length.toString()} files to add to the project`);
            for (const filePath of elmFilePaths) {
                connection.console.info(`Adding ${filePath.path.toString()}`);
                const fileContent = fs_1.readFileSync(filePath.path.toString(), "utf8");
                let tree;
                tree = parser.parse(fileContent);
                forest.setTree(vscode_uri_1.URI.file(filePath.path).toString(), filePath.writable, true, tree);
            }
            forest.treeIndex.forEach(item => {
                connection.console.info(`Adding imports ${item.uri.toString()}`);
                imports.updateImports(item.uri, item.tree, forest);
            });
            connection.console.info("Done parsing all files.");
        }
        catch (error) {
            connection.console.info(error.toString());
        }
    }
    findElmHome() {
        const elmHomeVar = process.env.ELM_HOME;
        if (elmHomeVar) {
            return elmHomeVar;
        }
        return utils.isWindows
            ? `${os_1.default.homedir()}/AppData/Roaming/elm`
            : `${os_1.default.homedir()}/.elm`;
    }
    findElmFilesInFolders(elmFolders) {
        let elmFilePaths = [];
        for (const element of elmFolders) {
            elmFilePaths = elmFilePaths.concat(this.findElmFilesInFolder(element));
        }
        return elmFilePaths;
    }
    findElmFilesInFolder(element) {
        return glob_1.default
            .sync(`${element.path}/**/*.elm`)
            .map(path => ({ path, writable: element.writable }));
    }
    registerProviders(connection, forest, elmWorkspace, imports, settings, parser) {
        this.initialize(connection, forest, elmWorkspace, imports, parser);
        const documentEvents = new documentEvents_1.DocumentEvents(connection);
        const textDocumentEvents = new textDocumentEvents_1.TextDocumentEvents(documentEvents);
        const documentFormatingProvider = new documentFormatingProvider_1.DocumentFormattingProvider(connection, elmWorkspace, textDocumentEvents, settings);
        const elmAnalyse = new elmAnalyseDiagnostics_1.ElmAnalyseDiagnostics(connection, elmWorkspace, textDocumentEvents, settings, documentFormatingProvider);
        // tslint:disable:no-unused-expression
        new astProvider_1.ASTProvider(connection, forest, documentEvents, imports, parser);
        new foldingProvider_1.FoldingRangeProvider(connection, forest);
        new completionProvider_1.CompletionProvider(connection, forest, imports);
        new hoverProvider_1.HoverProvider(connection, forest, imports);
        new diagnosticsProvider_1.DiagnosticsProvider(connection, elmWorkspace, textDocumentEvents, settings, elmAnalyse);
        new definitionProvider_1.DefinitionProvider(connection, forest, imports);
        new referencesProvider_1.ReferencesProvider(connection, forest, imports);
        new documentSymbolProvider_1.DocumentSymbolProvider(connection, forest);
        new workspaceSymbolProvider_1.WorkspaceSymbolProvider(connection, forest);
        new codeLensProvider_1.CodeLensProvider(connection, forest, imports);
        new renameProvider_1.RenameProvider(connection, forest, imports);
        new codeActionProvider_1.CodeActionProvider(connection, elmAnalyse);
    }
}
exports.Server = Server;
//# sourceMappingURL=server.js.map